#!/usr/bin/env python3

import setuptools


with open("README.md", "r") as fh:
	long_description = fh.read()


setuptools.setup(
	name="clirail",
	version="1.7.1",
	author="Midgard",
	author_email=None,
	description="Command line application for the iRail API",
	long_description=long_description,
	long_description_content_type="text/markdown",
	scripts=["bin/clirail"],

	url="https://framagit.org/Midgard/clirail",
	project_urls={
		"Source": "https://framagit.org/Midgard/clirail",
		"Change log": "https://framagit.org/Midgard/clirail/-/blob/master/CHANGELOG.md",
		"Bug tracker": "https://framagit.org/Midgard/clirail/-/issues",
	},

	classifiers=[
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
		"License :: OSI Approved :: ISC License (ISCL)",
		"Operating System :: OS Independent",
		"Natural Language :: English",
		"Natural Language :: Dutch",
		"Environment :: Console",
		"Development Status :: 5 - Production/Stable",
		"Intended Audience :: End Users/Desktop",
		"Topic :: Utilities",
	],

	packages=setuptools.find_packages(),
	python_requires=">=3.6",
	install_requires=[
		"requests",
		"python-dateutil",
	],
)



#!/usr/bin/env python

# clirail: command line interface to iRail
# Copyright © 2019 Midgard
#
# This program is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program.
# If not, see <http://www.gnu.org/licenses/>.


import sys
import json
from typing import Sequence, Mapping
import pathlib
from datetime import datetime, timedelta
from dateutil import parser as dateparser
import requests
from .data import station_codes
from .xdg import XDG_DATA_HOME
from .printing import print_connections, print_liveboard, print_disturbances
from .messages import t
from .config import LANG


BASE_URL = "https://api.irail.be"
LIVEBOARD_URL = f"{BASE_URL}/liveboard/"
CONNECTIONS_URL = f"{BASE_URL}/connections/"
DISTURBANCES_URL = f"{BASE_URL}/disturbances/"


def load_or_fetch_stations():
	stations_file = XDG_DATA_HOME / "clirail" / ("stations_" + LANG + ".json")
	stations_url = f"{BASE_URL}/stations/"

	try:
		with open(str(stations_file), mode="r") as file:
			return json.load(file)["station"]
	except FileNotFoundError:
		print(t("public_log"), file=sys.stderr)
		_ = input()
		print(t("downloading_stations"))
		print("")
		stations_file.parent.mkdir(exist_ok=True, parents=True)
		req = requests.get(stations_url, params={"format": "json", "lang": LANG})
		stations = req.json()

		with open(str(stations_file), mode="w") as file:
			json.dump(stations, file)

		return stations["station"]

STATIONS: Sequence[Mapping] = load_or_fetch_stations()


def uri_to_id(uri, default=None):
	for station in STATIONS:
		if station["@id"] == uri:
			return station["id"]
	return default


def tele_to_uri(telegraphic_code, default=None):
	return station_codes.get(str.upper(telegraphic_code), default)

def uri_to_tele(in_uri, default=None):
	for telegraphic_code, uri in station_codes.items():
		if uri == in_uri:
			return telegraphic_code
	return default


def irail_date(moment):
	"""
	Convert a datetime to the "ddmmyy" and "HHMM" that iRail expects.
	"""
	if moment is None:
		return None, None
	else:
		return moment.strftime("%d%m%y"), moment.strftime("%H%M")


def get_liveboard(station=None, station_is_id=True, moment=None, departure=True, alerts=False):
	date, time = irail_date(moment)

	params = {
		"lang": LANG,
		"format": "json",

		("id" if station_is_id else "station"): station,
		"alerts": alerts,
		"arrdep": "departure" if departure else "arrival",
		"date": date, "time": time
	}

	req = requests.get(LIVEBOARD_URL, params=params)
	return req.json()


def get_connections(fro, to, moment=None, alerts=False):
	date, time = irail_date(moment)

	params = {
		"lang": LANG,
		"format": "json",

		"from": fro, "to": to,
		"alerts": alerts,
		"date": date, "time": time
	}

	req = requests.get(CONNECTIONS_URL, params=params)
	return req.json()


def get_disturbances():
	params = {
		"lang": LANG,
		"format": "json"
	}

	req = requests.get(DISTURBANCES_URL, params=params)
	return req.json()


# FIXME split into get_summary and print_summary
def show_summary():
	now = datetime.now()
	hour_ago = now - timedelta(hours=1)

	for station in ["Brugge", "Gent-Sint-Pieters", "Brussel-Zuid", "Liège-Guillemins"]:
		canceled   = [0, 0]
		delay_6    = [0, 0]
		delay_30   = [0, 0]
		delay_more = [0, 0]
		on_time    = [0, 0]

		for departure in get_liveboard(station, False, hour_ago)["departures"]["departure"]:
			is_canceled = departure["canceled"] != "0"
			departure_time = int(departure["time"]) + (0 if is_canceled else int(departure["delay"]))
			past = departure_time < now.timestamp()
			delay = int(departure["delay"])

			if is_canceled:
				canceled[past] += 1
			elif delay == 0:
				on_time[past] += 1
			elif delay <= 6 * 60:
				delay_6[past] += 1
			elif delay <= 30 * 60:
				delay_30[past] += 1
			else:
				delay_more[past] += 1

		print("\x1b[1m" + station + "\x1b[0m")
		print(f"{t('delay_0'):>16}: {on_time[0]:3d} {on_time[1]:3d}")
		print(f"{t('delay_1'):>16}: {delay_6[0]:3d} {delay_6[1]:3d}")
		print(f"{t('delay_6'):>16}: {delay_30[0]:3d} {delay_30[1]:3d}")
		print(f"{t('delay_30'):>16}: {delay_more[0]:3d} {delay_more[1]:3d}")
		print(f"{t('delay_canceled'):>16}: {canceled[0]:3d} {canceled[1]:3d}\n")

	print(f"{t('departed'):>18}* ↑   ↑ {t('not_yet_departed')}")

	print(f"*{t('canceled_explanation')}")


def coalesce(iterable):
	"""
	Return first non-None element of iterable, or None if none are non-None.
	"""
	for element in iterable:
		if element is not None:
			return element
	return None


STATION_ID, STATION_FREEFORM = range(2)

def local_resolve_station(station_input):
	"""
	Try to locally resolve station. Returns a tuple: station description and the type of station
	description: either STATION_ID or STATION_FREEFORM.
	"""
	if station_input is None:
		return None, STATION_FREEFORM

	id_from_tele = uri_to_id(tele_to_uri(station_input))
	if id_from_tele is not None:
		return id_from_tele, STATION_ID

	return station_input, STATION_FREEFORM


def print_help():
	print(t("help"))


def main(fro=None, to=None, moment=None):
	if fro in ("-h", "--help"):
		print_help()
		return

	if not fro:
		show_summary()
		return

	if fro == "d":
		disturbances = get_disturbances()
		print_disturbances(disturbances)
		return

	fro, fro_type = local_resolve_station(fro)
	to,  to_type  = local_resolve_station(to)

	try:
		moment_datetime = dateparser.parse(moment) if moment else None
	except ValueError:
		print(t("didnt_understand_moment").format(moment=moment),
			file=sys.stderr)
		return

	if to:
		connections = get_connections(fro, to, moment=moment_datetime)
		if "error" in connections:
			# TODO Translate common messages
			print(connections["message"])
			return
		print_connections(connections)

	else:
		liveboard = get_liveboard(fro, fro_type == STATION_ID, moment=moment_datetime)
		if "error" in liveboard:
			print(liveboard["message"])
			return
		telegraphic_code = uri_to_tele(liveboard["stationinfo"]["@id"])
		print_liveboard(liveboard, telegraphic_code)



# clirail: command line interface to iRail
# Copyright © 2019 Midgard
# License: GPLv3+

from datetime import datetime, date
from .messages import t
import re
from itertools import chain


icons = [
	("BUS",  "🚌"), # bus, typically to replace train in case of railroad works
	("EUR",  "🚄"), # Eurostar
	("ICE",  "🚄"), # ICE, German high-speed train
	("THA",  "🚄"), # Thalys
	("TGV",  "🚄"), # TGV, French high-speed train
	("IC",   "🚅"), # intercity train
	("ICT",  "🚅"), # tourist intercity train
	("P",    "🚅"), # peak train
	("TRN",  "🚅"), # just a train
	("EXT",  "🚅"), # replacement train or something
	("S",    "🚂"), # sprinter train
	("L",    "🚂"), # local train
	("WALK", "🚶"), # walking
	("",     "🚀")  # unknown, go by rocket
]

IGNORED_ALERTS = {
	"Mondmaskerplicht",
	"Port du masque obligatoire",
	"Face mask mandatory",
	"Maskenpflicht"
}


def without_colours(s):
	return re.sub(r"\x1b\[[^m]*m", "", s)


def icon_from_ref(ref):
	for prefix, icon in icons:
		if ref.startswith(prefix):
			return icon
	assert False


def format_duration(seconds, hm=t("hm_duration_fmt"), m=t("m_duration_fmt")):
	minutes = int(seconds) // 60
	if minutes < 60:
		return m.format(minutes)
	return hm.format(minutes // 60, minutes % 60)

def format_delay(seconds):
	seconds = int(seconds)
	formt = "\x1b[33m" if seconds // 60 < 5 else "\x1b[91m"
	return "" if seconds == 0 else (
		" " + formt + sign_str(seconds) +
		format_duration(abs(seconds), t("hm_delay_fmt"), t("m_delay_fmt")) +
		"\x1b[0m"
	)


def format_time(timestamp):
	return datetime.fromtimestamp(int(timestamp)).strftime("%H:%M")


def format_datetime(timestamp):
	return datetime.fromtimestamp(int(timestamp)).strftime("%Y-%m-%d, %H:%M")


def sign_str(number):
	return "-" if number < 0 else "+"


def substring_from_last(string, char):
	try:
		last_pos = string.rindex(char)
		return string[last_pos+1:]
	except ValueError:
		return string


def print_liveboard_item(item):
	canceled = item["canceled"] != "0"
	delay_str = format_delay(int(item["delay"])) if not canceled else ""

	time_str = datetime.fromtimestamp(int(item["time"])).strftime("%H:%M")
	pretty_to = pad_trunc(item["station"] + " ", 30, ".")

	pretty_platform = pad_trunc(" " + item["platform"], 4, ".", left=False)

	ref = substring_from_last(item["vehicle"], ".")

	print((
		"{vehicle_id:>8} {icon} {to}.{pretty_platform}  {time_str}{delay_str}"
		if not canceled else
		"\x1b[9m{vehicle_id:>8} {icon} {to}.....  {time_str}\x1b[0m \x1b[31m{canceled_str}\x1b[0m"
	).format(
		**item,
		icon=icon_from_ref(ref),
		vehicle_id=ref,
		pretty_platform=pretty_platform,
		to=pretty_to,
		time_str=time_str,
		delay_str=delay_str,
		canceled_str=t("canceled")
	))


def print_liveboard(liveboard, tele=None):
	print()
	print(t("departure_in" if tele is None else "departure_in_w_code").format(
		station="\x1b[1m" + liveboard["station"] + "\x1b[0m",
		telegraphic_code=tele
	))
	print()

	print("\x1b[4;3m   {train:8} {destination:25} {platform:>9}  {departure:15}\x1b[0m".format(
		train=t("train"), destination=t("destination"), platform=t("platform"),
		departure=t("departure")
	))

	prev_date = date.today()
	for line in liveboard["departures"]["departure"]:
		dep_time = datetime.fromtimestamp(int(line["time"]))
		if dep_time.date() != prev_date:
			print(" " * 12 + format_date(dep_time))
			prev_date = dep_time.date()
		print_liveboard_item(line)


def print_connection_item(item, max_duration_len, max_delay_len):
	print()

	if item["departure"]["canceled"] != "0" or item["arrival"]["canceled"] != "0":
		print(t("canceled").upper())

	dep = item["departure"]
	arr = item["arrival"]
	dep_ref = substring_from_last(dep["vehicle"], ".")
	arr_ref = substring_from_last(arr["vehicle"], ".")

	print((
		(" " * max_duration_len) + " ┌← {time}{delay:>{max_delay_len}}  {fro}.{platform} {ref} {icon} {direction}"
	).format(
		max_delay_len=max_delay_len,
		fro=pad_trunc(dep["station"] + " ", 20, "."),
		platform=pad_trunc(" " + dep["platform"], 3, ".", left=False),
		ref=pad_trunc(" " + dep_ref, 10, ".", left=False),
		icon=icon_from_ref(dep_ref),
		time=format_time(dep["time"]),
		delay=format_delay(dep["delay"]),
		direction=dep["direction"]["name"]
	))

	if "vias" in item:
		for via in item["vias"]["via"]:
			via_dep = via["departure"]
			via_arr = via["arrival"]
			via_ref = substring_from_last(via_dep["vehicle"], ".")

			print((
				(" " * max_duration_len) + " └→ {time}{delay:>{max_delay_len}}\x1b[38;5;245m  {via}.{platform:2}\x1b[0m"
			).format(
				max_delay_len=max_delay_len,
				via=pad_trunc(via["station"] + " ", 20, "."),
				platform=pad_trunc(" " + via_arr["platform"], 3, ".", left=False),
				time=format_time(via_arr["time"]),
				delay=format_delay(via_arr["delay"])
			))

			if int(via_dep.get("walking")):
				print((
					(" " * max_duration_len) + " ┌→ {time}{delay:>{max_delay_len}}\x1b[38;5;245m  {via}. 🚶 {go_to_next_station}\x1b[0m"
				).format(
					max_delay_len=max_delay_len,
					via=pad_trunc(via["station"] + " ", 20, "."),
					time=format_time(via_arr["time"]),
					delay=format_delay(via_arr["delay"]),
					go_to_next_station=t("go_to_next_station")
				))
			else:
				print((
					(" " * max_duration_len) + " ┌← {time}{delay:>{max_delay_len}}\x1b[38;5;245m  {via}.{platform:2} {ref} {icon} {direction}\x1b[0m"
				).format(
					max_delay_len=max_delay_len,
					via=pad_trunc(via["station"] + " ", 20, "."),
					platform=pad_trunc(" " + via_dep["platform"], 3, ".", left=False),
					ref=pad_trunc(" " + via_ref, 10, ".", left=False),
					icon=icon_from_ref(via_ref),
					time=format_time(via_dep["time"]),
					delay=format_delay(via_dep["delay"]),
					direction=via_dep["direction"]["name"]
				))

	print((
		"{duration:>{max_duration_len}} └→ {time}{delay:>{max_delay_len}}  {to}.{platform:2} "
	).format(
		max_duration_len=max_duration_len,
		max_delay_len=max_delay_len,
		duration=format_duration(item["duration"]),
		to=pad_trunc(arr["station"] + " ", 20, "."),
		platform=pad_trunc(" " + arr["platform"], 3, ".", left=False),
		time=format_time(arr["time"]),
		delay=format_delay(arr["delay"])
	))


	for alert in item.get("alerts", {}).get("alert", []):
		if alert["header"] in IGNORED_ALERTS:
			continue
		print("""⚠  {}
\x1b[38;5;245m{}
van {} tot {}\x1b[0m""".format(alert["header"], alert["description"], format_datetime(alert["startTime"]), format_datetime(alert["endTime"])))

	remarks = set(remark["description"] for remark in item.get("remarks", {}).get("remark", []))
	for remark in remarks:
		print((" " * max_duration_len) + "  \x1b[38;5;245m{}\x1b[0m".format(remark))


def delay_str_len(delay):
	return len(without_colours(format_delay(delay)))


def max_delay_str_length(connection_item):
	return max(
		delay_str_len(connection_item["departure"]["delay"]),
		delay_str_len(connection_item["arrival"]["delay"]),
		*(
			delay_str_len(via["departure"]["delay"])
			for via in connection_item.get("vias", {}).get("via", [])
		)
	)


def format_date(d):
	return "\x1b[1m{:%Y-%m-%d} {}\x1b[0m".format(
		d,
		t("day_of_week_{}".format(d.weekday()))
	)


def print_connections(connections):
	max_duration_len = max(max((
		len(format_duration(item["duration"]))
		for item in connections["connection"]
	), default=0), len(t("duration")))

	max_delay_len = max(chain(
		(
			max_delay_str_length(item)
			for cs in connections["connection"]
			for item in cs.get("vias", {}).get("via", [])
		), (
			max_delay_str_length(item)
			for item in connections["connection"]
		)
	), default=0)

	print("\x1b[4;3m{duration:>{max_duration_len}} {dep_arr:>8} {delay_spaces} {station:12} {platform:>11} {train:>10}    {towards}\x1b[0m".format(
		max_duration_len=max_duration_len,
		duration=t("duration"), dep_arr=t("dep_arr"),
		delay_spaces=" " * max_delay_len,
		station=t("station"), platform=t("platform"),
		train=t("train"), towards=t("towards")
	))

	prev_date = date.today()
	for line in connections["connection"]:
		dep_time = datetime.fromtimestamp(int(line["departure"]["time"]))
		if dep_time.date() != prev_date:
			print("\n" + format_date(dep_time))
			prev_date = dep_time.date()

		print_connection_item(line, max_duration_len, max_delay_len)


def print_disturbances(disturbances):
	for disturbance in reversed(disturbances["disturbance"]):
		print("""{}
\x1b[38;5;245m{}\x1b[0m
\x1b[38;5;33m{}\x1b[0m
""".format(
	disturbance["title"],
	re.sub(
		r'(Webnews|Info) [A-Z]{2}',
		'',
		re.sub(r'[.:]([A-Z*-])', lambda m: ".\n" + m[1], disturbance["description"])
	).strip(),
	disturbance["link"]))


def pad_trunc(string, width, filler=" ", left=True):
	if len(string) >= width:
		return string[:width]
	if left:
		return (string + (filler * width))[:width]
	else:
		return ((filler * width) + string)[-width:]
